freeslot(
-- flower object
"MT_CDFLOWER",
"MT_CDFLOWERSPARKLE",
-- flower state
"S_CDFLOWERWAIT",
"S_CDFLOWERFLY",
"S_CDFLOWERSPROUT",
"S_CDFLOWERSPARKLE",
-- flower animations
"SPR_CDSE",
"SPR_CDSP",
"SPR_CDEX",
"SPR_CDP3",
"SPR_CDP4",
"SPR_CDR5"
)

local msToTics = function(ms)
	return FixedInt(FixedMul(FixedDiv(ms*FRACUNIT, 1000*FRACUNIT), TICRATE*FRACUNIT))
end
local ticsToMs = function(tics)
	return FixedInt(FixedMul(FixedDiv(tics*FRACUNIT, TICRATE*FRACUNIT), 1000*FRACUNIT))
end

local mimicDeathFling = function(mo)
	local target = mo.target

	local flingSpeed = FixedHypot(target.momx, target.momy)
	local flingAngle = 0

	mo.momx = 0
	mo.momy = 0
	mo.momz = 0

	flingAngle = target.angle

	if (P_RandomByte(PR_ITEM_RINGS) & 1) then
		flingAngle = flingAngle - ANGLE_45/2
	else
		flingAngle = flingAngle + ANGLE_45/2
	end

	--local inflictorSpeed = 0
	--if (!P_MobjWasRemoved(inflictor)) then
	--	inflictorSpeed = FixedHypot(inflictor.momx, inflictor.momy)
	--	if (inflictorSpeed > flingSpeed) then
	--		flingSpeed = inflictorSpeed
	--	end
	--end

	local battle = (gametyperules & (GTR_BUMPERS | GTR_BOSS)) == GTR_BUMPERS
	local thrust_divisor = 0
	if battle then
		thrust_divisor = 1
	else
		thrust_divisor = 3
	end
	P_InstaThrust(mo, flingAngle, max(flingSpeed, 6 * target.scale) / thrust_divisor)
	local momz = 0
	if battle then
		--momz = 20*FRACUNIT 
		momz = 16*FRACUNIT
	else
		--momz = 18*FRACUNIT 
		momz = 14*FRACUNIT
	end
	P_SetObjectMomZ(mo, momz, false)
end

-- seed has normal timing in the air (part of death sprites)
-- on contact with ground switch to CDEX
-- CDEX goes A B A B A B (60ms difference for each)
-- flower sprite goes:
-- A for 160ms
-- then infinitely alternate between B and C for 200ms

states[S_CDFLOWERWAIT] = {
	sprite = SPR_NULL,
	tics = -1,
}

states[S_CDFLOWERFLY] = {
	sprite = SPR_CDSE,
	frame = A,
	tics = -1,
	var1 = 3,
	var2 = 3,
}

states[S_CDFLOWERSPROUT] = {
	sprite = SPR_CDEX,
	frame = A,
	tics = -1,
}

states[S_CDFLOWERSPARKLE] = {
	sprite = SPR_CDSP,
	--sprite = SPR_SPBM,
	frame = A|FF_ANIMATE,
	var1 = 3,
	var2 = 3,
	tics = 4*3,
	nextstate = S_NULL,
}

mobjinfo[MT_CDFLOWERSPARKLE] = {
	--flags = MF_NOGRAVITY|MF_NOCLIPHEIGHT|MF_NOCLIP,
	flags = MF_NOSQUISH|MF_NOGRAVITY|MF_NOCLIPHEIGHT,
	doomednum = -1,
	spawnstate = S_CDFLOWERSPARKLE,
	deathstate = S_NULL,
}

mobjinfo[MT_CDFLOWER] = {
	--flags = MF_NOGRAVITY|MF_NOCLIPHEIGHT|MF_NOCLIP,
	flags = MF_NOSQUISH|MF_APPLYTERRAIN,
	doomednum = -1,
	spawnstate = S_CDFLOWERWAIT,
	deathstate = S_NULL,
}

local P_Is_ObjectOnGround = function(mo)
	if (mo.eflags & MFE_VERTICALFLIP) then
		if (mo.z+mo.height >= mo.ceilingz) then
			return true
		end
	else
		if (mo.z <= mo.floorz) then
			return true
		end
	end
	return false
end

local getCDFlowerSprite = function(mo)
	if mo.skin == "tagataga" then
		return SPR_CDP4
	end
	if mo.skin == "kamakama" then
		return SPR_CDP3
	end
	if mo.skin == "noronoro" then
		return SPR_CDR5
	end
	return SPR_UNKN
end

addHook("MobjDeath", function (mo)
--addHook("MobjThinker", function(mo)
		if not (mo and mo.valid and mo.player) then return false end
		
		if mo.skin == "tagataga" or mo.skin == "kamakama" or mo.skin == "noronoro" then
			local cmd = mo.player.cmd
			
			/*if (cmd.buttons & BT_LUAA) == 0 then
				return
			end*/
			
			-- for some reason, spawning custom mobj types doesn't account for
			-- the current floor/ceiling of the parent object. we spawn an
			-- existing basegame object instead first to fix it, then
			-- immediately switch its type to our custom one
			local seed = P_SpawnMobjFromMobj(mo, 0, 0, 35*FRACUNIT, MT_DRIFTCLIP)
			seed.type = MT_CDFLOWER
			seed.state = S_CDFLOWERWAIT
			seed.sprite = SPR_NULL
			seed.scale = mo.scale*2+(FRACUNIT/8)
			seed.radius = mo.radius
			seed.reactiontime = 0
			seed.target = mo
			seed.health = mo.hitlag
			seed.color = mo.player.skincolor
			seed.cusval = getCDFlowerSprite(mo)
			seed.flags = seed.flags | MF_NOGRAVITY
		end
		return false
end, MT_PLAYER)
	
addHook("MobjThinker", function(mo)
	return false
end, MT_CDFLOWERSPARKLE)

addHook("MobjThinker", function(mo)
	if not (mo and mo.valid and mo.target) then return false end
	if mo.state ~= S_CDFLOWERWAIT and mo.target then
		mo.target.sprite = SPR_NULL
	end
	if P_Is_ObjectOnGround(mo) and (mo.state ~= S_CDFLOWERSPROUT) then
		if mo.health >= 1000 then
			mo.momx = 0
			mo.momy = 0
			mo.momz = 0
			mo.flags = mo.flags | MF_NOGRAVITY
			mo.z = P_FloorzAtPos(mo.x, mo.y, mo.z + 100*FRACUNIT, mo.height)
			mo.state = S_CDFLOWERSPROUT
			mo.health = 0
			mo.scale = mo.extravalue1
			return false
		end
	else
		if mo.state == S_CDFLOWERFLY then
		if (mo.health % 3) == 0 then
			local sparkle = P_SpawnMobjFromMobj(mo, mo.momx, mo.momy, mo.momz, MT_CDFLOWERSPARKLE)
		end
		mo.health = mo.health + ticsToMs(1)
		end
	end
	if mo.state == S_CDFLOWERWAIT then
		if mo.reactiontime == 0 then
			mo.health = mo.target.hitlag
			if mo.health == 0 then
				-- mo.sprite = SPR_CDSE
				mo.reactiontime  = 1
				return false
			end
		end
		if mo.reactiontime == 1 then
			mo.health = mo.target.hitlag
			if mo.health == 0 then
				mo.reactiontime  = 2
				return false
			end
		end
		if mo.reactiontime == 2 then
			mo.health = mo.target.hitlag
			if mo.health == 1 then
				mo.frame = A
				mo.state = S_CDFLOWERFLY
				mo.flags = mo.flags & ~MF_NOGRAVITY
				mo.sprite = SPR_CDSE
				mo.shadowscale = mo.target.shadowscale
				mo.whiteshadow = mo.target.whiteshadow
				mo.shadowcolor = mo.target.shadowcolor
				mo.extravalue1 = mo.target.scale+FRACUNIT/8
				mimicDeathFling(mo)
				return false
			end
		end
	end
	if mo.state == S_CDFLOWERSPROUT then
		if mo.health > 60 then
			mo.sprite = SPR_CDEX
			mo.frame = A
		end
		if mo.health > 120 then
			mo.sprite = SPR_CDEX
			mo.frame = B
		end
		if mo.health > 180 then
			mo.sprite = SPR_CDEX
			mo.frame = A
		end
		if mo.health > 240 then
			mo.sprite = SPR_CDEX
			mo.frame = B
		end
		if mo.health > 300 then
			mo.sprite = SPR_CDEX
			mo.frame = A
		end
		if mo.health > 360 then
			mo.sprite = SPR_CDEX
			mo.frame = B
		end
		if mo.health > 420 then
			mo.sprite = mo.cusval
			mo.frame = A
		end
		if mo.health > 580 then
			mo.sprite = mo.cusval
			mo.frame = (((mo.health - 580) % 400) / 200) + 1
		end
		mo.health = mo.health + ticsToMs(1)
	end
	return false
end, MT_CDFLOWER)